@@ -24,6 +24,11 @@
 #define POR_LINE_LENGTH         80
 #define POR_LABEL_NAME_PREFIX   "labels"
 
+#define MAX_FORMAT_TYPE       120
+#define MAX_FORMAT_WIDTH      100
+#define MAX_FORMAT_DECIMALS   100
+#define MAX_STRING_LENGTH   20000
+
 #define MAX_VARS    1000000
 #define MAX_WIDTH   1000000
 #define MAX_LINES   1000000
@@ -35,7 +40,7 @@ static readstat_error_t read_string(por_ctx_t *ctx, char *data, size_t len);
 
 static readstat_error_t por_update_progress(por_ctx_t *ctx) {
     readstat_io_t *io = ctx->io;
-    return io->update(ctx->file_size, ctx->progress_handler, ctx->user_ctx, io->io_ctx);
+    return io->update(ctx->file_size, ctx->handle.progress, ctx->user_ctx, io->io_ctx);
 }
 
 static ssize_t read_bytes(por_ctx_t *ctx, void *dst, size_t len) {
@@ -119,21 +124,21 @@ static readstat_error_t read_double_with_peek(por_ctx_t *ctx, double *out_double
 
     len = por_utf8_encode(buffer, i, utf8_buffer, sizeof(utf8_buffer), ctx->byte2unicode);
     if (len == -1) {
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(error_buf, sizeof(error_buf), "Error converting double string (length=%" PRId64 "): %.*s", 
                     i, (int)i, buffer);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_CONVERT;
         goto cleanup;
     }
     
-    bytes_read = readstat_por_parse_double(utf8_buffer, len, &value, ctx->error_handler, ctx->user_ctx);
+    bytes_read = readstat_por_parse_double(utf8_buffer, len, &value, ctx->handle.error, ctx->user_ctx);
     if (bytes_read == -1) {
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(error_buf, sizeof(error_buf), "Error parsing double string (length=%" PRId64 "): %.*s [%s]", 
                     len, (int)len, utf8_buffer, buffer);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
@@ -200,7 +205,7 @@ static readstat_error_t maybe_read_string(por_ctx_t *ctx, char *data, size_t len
         return retval;
     }
     
-    if (value <= 0 || value > 20000 || isnan(value)) {
+    if (value < 0 || value > MAX_STRING_LENGTH || isnan(value)) {
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
     }
@@ -218,10 +223,10 @@ static readstat_error_t maybe_read_string(por_ctx_t *ctx, char *data, size_t len
     size_t bytes_encoded = por_utf8_encode(ctx->string_buffer, string_length, 
             data, len - 1, ctx->byte2unicode);
     if (bytes_encoded == -1) {
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(error_buf, sizeof(error_buf), "Error converting string: %.*s", 
                     (int)string_length, ctx->string_buffer);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_CONVERT;
         goto cleanup;
@@ -261,12 +266,22 @@ static readstat_error_t read_variable_count_record(por_ctx_t *ctx) {
         retval = READSTAT_ERROR_MALLOC;
         goto cleanup;
     }
-    if (ctx->info_handler) {
-        if (ctx->info_handler(-1, ctx->var_count, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+
+    if (ctx->handle.metadata) {
+        readstat_metadata_t metadata = {
+            .row_count = -1,
+            .var_count = ctx->var_count,
+            .creation_time = ctx->timestamp,
+            .modified_time = ctx->timestamp,
+            .file_format_version = ctx->version,
+            .file_label = ctx->file_label
+        };
+        if (ctx->handle.metadata(&metadata, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
     }
+        
 cleanup:
     return retval;
 }
@@ -318,17 +333,17 @@ static readstat_error_t read_variable_record(por_ctx_t *ctx) {
 
     for (i=0; i<sizeof(formats)/sizeof(spss_format_t *); i++) {
         spss_format_t *format = formats[i];
-        if ((retval = read_integer_in_range(ctx, 0, 100, &value)) != READSTAT_OK) {
+        if ((retval = read_integer_in_range(ctx, 0, MAX_FORMAT_TYPE, &value)) != READSTAT_OK) {
             goto cleanup;
         }
         format->type = value;
 
-        if ((retval = read_integer_in_range(ctx, 0, 100, &value)) != READSTAT_OK) {
+        if ((retval = read_integer_in_range(ctx, 0, MAX_FORMAT_WIDTH, &value)) != READSTAT_OK) {
             goto cleanup;
         }
         format->width = value;
 
-        if ((retval = read_integer_in_range(ctx, 0, 100, &value)) != READSTAT_OK) {
+        if ((retval = read_integer_in_range(ctx, 0, MAX_FORMAT_DECIMALS, &value)) != READSTAT_OK) {
             goto cleanup;
         }
         format->decimal_places = value;
@@ -340,8 +355,6 @@ cleanup:
 
 static readstat_error_t read_missing_value_record(por_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
-    double value;
-    char string[256];
     spss_varinfo_t *varinfo = NULL;
 
     if (ctx->var_offset < 0 || ctx->var_offset >= ctx->var_count) {
@@ -351,27 +364,27 @@ static readstat_error_t read_missing_value_record(por_ctx_t *ctx) {
     varinfo = &ctx->varinfo[ctx->var_offset];
 
     if (varinfo->type == READSTAT_TYPE_DOUBLE) {
-        if ((retval = read_double(ctx, &value)) != READSTAT_OK) {
-            goto cleanup;
-        }
-        varinfo->missing_values[varinfo->n_missing_values++] = value;
-        if (varinfo->n_missing_values > 3) {
-            retval = READSTAT_ERROR_PARSE;
+        if ((retval = read_double(ctx, &varinfo->missing_double_values[varinfo->n_missing_values])) != READSTAT_OK) {
             goto cleanup;
         }
     } else {
-        if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
+        if ((retval = read_string(ctx, varinfo->missing_string_values[varinfo->n_missing_values],
+                        sizeof(varinfo->missing_string_values[varinfo->n_missing_values]))) != READSTAT_OK) {
             goto cleanup;
         }
     }
+    if (varinfo->n_missing_values > 2) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
+    varinfo->n_missing_values++;
+
 cleanup:
     return retval;
 }
 
 static readstat_error_t read_missing_value_range_record(por_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
-    double value;
-    char string[256];
     spss_varinfo_t *varinfo = NULL;
 
     if (ctx->var_offset < 0 || ctx->var_offset == ctx->var_count) {
@@ -380,22 +393,22 @@ static readstat_error_t read_missing_value_range_record(por_ctx_t *ctx) {
     }
     varinfo = &ctx->varinfo[ctx->var_offset];
 
+    varinfo->missing_range = 1;
+    varinfo->n_missing_values = 2;
     if (varinfo->type == READSTAT_TYPE_DOUBLE) {
-        varinfo->missing_range = 1;
-        if ((retval = read_double(ctx, &value)) != READSTAT_OK) {
+        if ((retval = read_double(ctx, &varinfo->missing_double_values[0])) != READSTAT_OK) {
             goto cleanup;
         }
-        varinfo->missing_values[0] = value;
-        if ((retval = read_double(ctx, &value)) != READSTAT_OK) {
+        if ((retval = read_double(ctx, &varinfo->missing_double_values[1])) != READSTAT_OK) {
             goto cleanup;
         }
-        varinfo->missing_values[1] = value;
-        varinfo->n_missing_values = 2;
     } else {
-        if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
+        if ((retval = read_string(ctx, varinfo->missing_string_values[0],
+                        sizeof(varinfo->missing_string_values[0]))) != READSTAT_OK) {
             goto cleanup;
         }
-        if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
+        if ((retval = read_string(ctx, varinfo->missing_string_values[1],
+                        sizeof(varinfo->missing_string_values[1]))) != READSTAT_OK) {
             goto cleanup;
         }
     }
@@ -405,8 +418,6 @@ cleanup:
 
 static readstat_error_t read_missing_value_lo_range_record(por_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
-    double value;
-    char string[256];
     spss_varinfo_t *varinfo = NULL;
 
     if (ctx->var_offset < 0 || ctx->var_offset == ctx->var_count) {
@@ -415,16 +426,17 @@ static readstat_error_t read_missing_value_lo_range_record(por_ctx_t *ctx) {
     }
     varinfo = &ctx->varinfo[ctx->var_offset];
 
+    varinfo->missing_range = 1;
+    varinfo->n_missing_values = 2;
     if (varinfo->type == READSTAT_TYPE_DOUBLE) {
-        varinfo->missing_range = 1;
-        if ((retval = read_double(ctx, &value)) != READSTAT_OK) {
+        varinfo->missing_double_values[0] = -HUGE_VAL;
+        if ((retval = read_double(ctx, &varinfo->missing_double_values[1])) != READSTAT_OK) {
             goto cleanup;
         }
-        varinfo->missing_values[0] = -HUGE_VAL;
-        varinfo->missing_values[1] = value;
-        varinfo->n_missing_values = 2;
     } else {
-        if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
+        varinfo->missing_string_values[0][0] = '\0';
+        if ((retval = read_string(ctx, varinfo->missing_string_values[1],
+                        sizeof(varinfo->missing_string_values[1]))) != READSTAT_OK) {
             goto cleanup;
         }
     }
@@ -434,8 +446,6 @@ cleanup:
 
 static readstat_error_t read_missing_value_hi_range_record(por_ctx_t *ctx) {
     readstat_error_t retval = READSTAT_OK;
-    double value;
-    char string[256];
     spss_varinfo_t *varinfo = NULL;
 
     if (ctx->var_offset < 0 || ctx->var_offset == ctx->var_count) {
@@ -444,18 +454,19 @@ static readstat_error_t read_missing_value_hi_range_record(por_ctx_t *ctx) {
     }
     varinfo = &ctx->varinfo[ctx->var_offset];
 
+    varinfo->missing_range = 1;
+    varinfo->n_missing_values = 2;
     if (varinfo->type == READSTAT_TYPE_DOUBLE) {
-        varinfo->missing_range = 1;
-        if ((retval = read_double(ctx, &value)) != READSTAT_OK) {
+        if ((retval = read_double(ctx, &varinfo->missing_double_values[0])) != READSTAT_OK) {
             goto cleanup;
         }
-        varinfo->missing_values[0] = value;
-        varinfo->missing_values[1] = HUGE_VAL;
-        varinfo->n_missing_values = 2;
+        varinfo->missing_double_values[1] = HUGE_VAL;
     } else {
-        if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
+        if ((retval = read_string(ctx, varinfo->missing_string_values[0],
+                        sizeof(varinfo->missing_string_values[0]))) != READSTAT_OK) {
             goto cleanup;
         }
+        varinfo->missing_string_values[1][0] = '\0';
     }
 cleanup:
     return retval;
@@ -473,8 +484,8 @@ static readstat_error_t read_document_record(por_ctx_t *ctx) {
         if ((retval = read_string(ctx, string, sizeof(string))) != READSTAT_OK) {
             goto cleanup;
         }
-        if (ctx->note_handler) {
-            if (ctx->note_handler(i, string, ctx->user_ctx) != READSTAT_OK) {
+        if (ctx->handle.note) {
+            if (ctx->handle.note(i, string, ctx->user_ctx) != READSTAT_OK) {
                 retval = READSTAT_ERROR_USER_ABORT;
                 goto cleanup;
             }
@@ -499,7 +510,7 @@ static readstat_error_t read_variable_label_record(por_ctx_t *ctx) {
         goto cleanup;
     }
 
-    varinfo->label = malloc(strlen(string) + 1);
+    varinfo->label = realloc(varinfo->label, strlen(string) + 1);
     strcpy(varinfo->label, string);
 
 cleanup:
@@ -551,9 +562,11 @@ static readstat_error_t read_value_label_record(por_ctx_t *ctx) {
             }
             value.v.double_value = dval;
         }
-        if (ctx->value_label_handler(label_name_buf, value, label_buf, ctx->user_ctx) != READSTAT_HANDLER_OK) {
-            retval = READSTAT_ERROR_USER_ABORT;
-            goto cleanup;
+        if (ctx->handle.value_label) {
+            if (ctx->handle.value_label(label_name_buf, value, label_buf, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+                retval = READSTAT_ERROR_USER_ABORT;
+                goto cleanup;
+            }
         }
     }
     ctx->labels_offset++;
@@ -581,10 +594,10 @@ static readstat_error_t read_por_file_data(por_ctx_t *ctx) {
             if (info->type == READSTAT_TYPE_STRING) {
                 rs_retval = maybe_read_string(ctx, input_string, sizeof(input_string), &finished);
                 if (rs_retval != READSTAT_OK) {
-                    if (ctx->error_handler) {
+                    if (ctx->handle.error) {
                         snprintf(error_buf, sizeof(error_buf), "Error in %s (row=%d)", 
                                 info->name, ctx->obs_count+1);
-                        ctx->error_handler(error_buf, ctx->user_ctx);
+                        ctx->handle.error(error_buf, ctx->user_ctx);
                     }
                     goto cleanup;
                 } else if (finished) {
@@ -601,10 +614,10 @@ static readstat_error_t read_por_file_data(por_ctx_t *ctx) {
             } else if (info->type == READSTAT_TYPE_DOUBLE) {
                 rs_retval = maybe_read_double(ctx, &value.v.double_value, &finished);
                 if (rs_retval != READSTAT_OK) {
-                    if (ctx->error_handler) {
+                    if (ctx->handle.error) {
                         snprintf(error_buf, sizeof(error_buf), "Error in %s (row=%d)", 
                                 info->name, ctx->obs_count+1);
-                        ctx->error_handler(error_buf, ctx->user_ctx);
+                        ctx->handle.error(error_buf, ctx->user_ctx);
                     }
                     goto cleanup;
                 } else if (finished) {
@@ -614,8 +627,8 @@ static readstat_error_t read_por_file_data(por_ctx_t *ctx) {
                 }
                 value.is_system_missing = isnan(value.v.double_value);
             }
-            if (ctx->value_handler && !ctx->variables[i]->skip) {
-                if (ctx->value_handler(ctx->obs_count, ctx->variables[i], value, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+            if (ctx->handle.value && !ctx->variables[i]->skip) {
+                if (ctx->handle.value(ctx->obs_count, ctx->variables[i], value, ctx->user_ctx) != READSTAT_HANDLER_OK) {
                     rs_retval = READSTAT_ERROR_USER_ABORT;
                     goto cleanup;
                 }
@@ -685,8 +698,8 @@ readstat_error_t handle_variables(por_ctx_t *ctx) {
 
         int cb_retval = READSTAT_HANDLER_OK;
 
-        if (ctx->variable_handler) {
-            cb_retval = ctx->variable_handler(i, ctx->variables[i],
+        if (ctx->handle.variable) {
+            cb_retval = ctx->handle.variable(i, ctx->variables[i],
                     info->labels_index == -1 ? NULL : label_name_buf,
                     ctx->user_ctx);
         }
@@ -702,11 +715,11 @@ readstat_error_t handle_variables(por_ctx_t *ctx) {
             index_after_skipping++;
         }
     }
-    if (ctx->fweight_handler && ctx->fweight_name[0]) {
+    if (ctx->handle.fweight && ctx->fweight_name[0]) {
         for (i=0; i<ctx->var_count; i++) {
             spss_varinfo_t *info = &ctx->varinfo[i];
             if (strcmp(info->name, ctx->fweight_name) == 0) {
-                if (ctx->fweight_handler(ctx->variables[i], ctx->user_ctx) != READSTAT_HANDLER_OK) {
+                if (ctx->handle.fweight(ctx->variables[i], ctx->user_ctx) != READSTAT_HANDLER_OK) {
                     retval = READSTAT_ERROR_USER_ABORT;
                     goto cleanup;
                 }
@@ -723,20 +736,11 @@ readstat_error_t readstat_parse_por(readstat_parser_t *parser, const char *path,
     readstat_io_t *io = parser->io;
     unsigned char reverse_lookup[256];
     char vanity[5][40];
-    char file_label[21];
     char error_buf[1024];
 
     por_ctx_t *ctx = por_ctx_init();
     
-    ctx->info_handler = parser->info_handler;
-    ctx->metadata_handler = parser->metadata_handler;
-    ctx->note_handler = parser->note_handler;
-    ctx->fweight_handler = parser->fweight_handler;
-    ctx->variable_handler = parser->variable_handler;
-    ctx->value_handler = parser->value_handler;
-    ctx->value_label_handler = parser->value_label_handler;
-    ctx->error_handler = parser->error_handler;
-    ctx->progress_handler = parser->progress_handler;
+    ctx->handle = parser->handlers;
     ctx->user_ctx = user_ctx;
     ctx->io = io;
     ctx->row_limit = parser->row_limit;
@@ -772,7 +776,9 @@ readstat_error_t readstat_parse_por(readstat_parser_t *parser, const char *path,
         goto cleanup;
     }
 
-    readstat_convert(file_label, sizeof(file_label), vanity[1] + 20, 20, NULL);
+    retval = readstat_convert(ctx->file_label, sizeof(ctx->file_label), vanity[1] + 20, 20, NULL);
+    if (retval != READSTAT_OK)
+        goto cleanup;
     
     if (read_bytes(ctx, reverse_lookup, sizeof(reverse_lookup)) != sizeof(reverse_lookup)) {
         retval = READSTAT_ERROR_READ;
@@ -802,9 +808,9 @@ readstat_error_t readstat_parse_por(readstat_parser_t *parser, const char *path,
     }
     
     if (por_utf8_encode(check, sizeof(check), tr_check, sizeof(tr_check), ctx->byte2unicode) == -1) {
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(error_buf, sizeof(error_buf), "Error converting check string: %.*s", (int)sizeof(check), check);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_CONVERT;
         goto cleanup;
@@ -823,13 +829,6 @@ readstat_error_t readstat_parse_por(readstat_parser_t *parser, const char *path,
     if (retval != READSTAT_OK)
         goto cleanup;
 
-    if (ctx->metadata_handler) {
-        if (ctx->metadata_handler(file_label, NULL, ctx->timestamp, ctx->version, ctx->user_ctx) != READSTAT_HANDLER_OK) {
-            retval = READSTAT_ERROR_USER_ABORT;
-            goto cleanup;
-        }
-    }
-        
     while (1) {
         uint16_t tr_tag = read_tag(ctx);
         switch (tr_tag) {
@@ -881,7 +880,7 @@ readstat_error_t readstat_parse_por(readstat_parser_t *parser, const char *path,
                 if (retval != READSTAT_OK)
                     goto cleanup;
 
-                if (ctx->value_handler) {
+                if (ctx->handle.value) {
                     retval = read_por_file_data(ctx);
                 }
                 goto cleanup;
